package org.fictus

import scala.collection.mutable.ListBuffer
import org.fictus.internal.Method
import scala.collection.mutable.HashMap

class ExceptionWrapper(e: Exception) extends RuntimeException(e)

case class FictusResult[A](value: A)

trait Fictus extends MockProxy with Matchers with Assertions with Logging {
    implicit val context = new MockContext(this)

    def mocks2[M1](implicit context: MockContext, m1: Manifest[M1]): (M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"))
    }

    def mocks3[M1](implicit context: MockContext, m1: Manifest[M1]): (M1, M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"), mock(nameBase + "._3"))
    }

    def mock[M1, M2](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2]): (M1, M2) = (mock[M1], mock[M2])
    def mock[M1, M2, M3](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3]): (M1, M2, M3) = (mock[M1], mock[M2], mock[M3])
    def mock[M1, M2, M3, M4](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3], m4: Manifest[M4]): (M1, M2, M3, M4) = (mock[M1], mock[M2], mock[M3], mock[M4])

    /** @return a mock of the given type with a name of the given type. */
    def mock[A](implicit context: MockContext, manifest: Manifest[A]): A = mock(manifest.erasure.getSimpleName)

    /** @return a mock of the given type with the passed in name. */
    def mock[A](name: String)(implicit context: MockContext, manifest: Manifest[A]): A = {
        val mock = new Mock(name)
        createProxy(mock)
    }

    def missingBehaviors(expectation: Expectation[_]) = throw new PebkacError("\nExpected method is missing behaviors (returns or throws)\n" +
        expectation + " should look something like:\n" +
        "\t" + expectation + " --> <some return> OR\n" +
        "\t" + expectation + " returns <some return> OR\n" +
        "\t" + expectation + " throws <some exception>")

    def defaults(theDefaults: => Unit)(implicit context: MockContext) {
        def clearLastExpectation() = context.lastExpectation.isComplete match {
            case Some(expectation) => missingBehaviors(expectation)
            case None              => context.lastExpectation.clear
        }

        context.state.default
        clearLastExpectation()
        theDefaults
        clearLastExpectation()
        context.state.exitDefault
    }

    implicit def expect[A](expected: => A)(implicit context: MockContext): Expectation[A] = {
        context.lastExpectation.isComplete match {
            case Some(expectation) => missingBehaviors(expectation)
            case None =>
                expected
                context.lastExpectation.last[A]().get
        }
    }

    def test[A](unitFunction: => A)(implicit context: MockContext): A = {
        def checkLastexpectationDefined(implicit context: MockContext) =
            context.lastExpectation.isComplete match {
                case Some(expectation) => missingBehaviors(expectation)
                case None              =>
            }
        checkLastexpectationDefined

        context.state.replay

        try {
            val result = unitFunction
            context.expectations.verifyAllExpectationsAreConsumed
            context.state.done
            result
        } catch {
            case e: ExceptionWrapper => throw e.getCause
        }
    }
}

class Mock(name: String)(implicit context: MockContext) extends Invoke(name) {

    override val toString = name

    override def replay(method: Method, actualArgs: List[AnyRef]): AnyRef = {
        context.state.replay
        context.expectations.find(this, method, actualArgs) match {
            case Some(expectation) => replay(expectation, method, actualArgs)
            case None              => throw new PebkacError("Unexpected method called: " + name + "." + method.name + actualArgs.mkString("(", ", ", ")"))
        }
    }

    private def replay(expectation: RecordedBehavior, method: Method, actualArgs: List[AnyRef]) = {
        val matchers = expectation.expectation.matchers

        val argsMatch = matchers.zip(actualArgs).map {
            case (matcher, actual) => matcher.matches(actual)
        }

        context.logging.printSystemError(
            "REPLAY " + method.fullName +
                "\n  Matchers=" + matchers +
                "\n  Actual args=" + actualArgs +
                "\n  Behavior=" + expectation.behavior +
                "\n  Args Match=" + argsMatch)

        actualArgs.foreach { a =>
            context.logging.printSystemError("ACTUAL CLASS=" + Option(a).getOrElse("null").getClass)
        }

        if (!argsMatch.isEmpty && argsMatch.contains(false)) {

            case class Match(matcher: Matcher, actualArg: AnyRef, matches: Boolean)

            val matches = argsMatch.zip(matchers).zip(actualArgs).map { case ((matches, matcher), actualArg) => Match(matcher, actualArg, matches) }

            val expectedList = matches.map(m => m.matches match {
                case true  => m.matcher.expectedMsg
                case false => "->" + m.matcher.expectedMsg + "<-"
            })

            val actualList = matches.map(m => m.matches match {
                case true  => m.matcher.actualMsg(m.actualArg)
                case false => "->" + m.matcher.actualMsg(m.actualArg) + "<-"
            })

            val msg = "Expected arguments did not match the actual arguments!\n" +
                "expected: " + name + "." + method.name + expectedList.mkString("(", ", ", ")") + "\n" +
                "actual:   " + name + "." + method.name + actualList.mkString("(", ", ", ")")
            throw new ExpectationError(msg)
        }

        expectation.behavior.handle match {
            case ref: AnyRef => ref
            case null        => null
            case value => value match {
                case i: Boolean => java.lang.Boolean.valueOf(i)
                case i: Byte    => java.lang.Byte.valueOf(i)
                case i: Char    => java.lang.Character.valueOf(i)
                case i: Short   => java.lang.Short.valueOf(i)
                case i: Int     => java.lang.Integer.valueOf(i)
                case i: Long    => java.lang.Long.valueOf(i)
                case i: Float   => java.lang.Float.valueOf(i)
                case i: Double  => java.lang.Double.valueOf(i)
            }
        }
    }

    override def record(method: Method, args: List[AnyRef]): AnyRef = {
        context.lastExpectation.store(Expectation(this, method, context.currentMatchers.getAndClear, args, "ugh"))
        Types.defaultValue(method.returnType)
    }
}

case class RecordedBehavior(expectation: Expectation[_], behavior: Behavior) {
    val methodName = expectation.methodName + behavior
}

private[fictus] class RecordedMatchers(context: MockContext) {
    private val matchers = new ListBuffer[Matcher]

    def append(matcher: Matcher) {
        matchers.append(matcher)
    }

    def getAndClear: List[Matcher] = {
        val theMatchers = matchers.toList
        matchers.clear
        theMatchers
    }
}

trait Expectations {
    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A]
    def lastOption(): Option[RecordedBehavior]
    def convertLastBehaviorToAnyTimes(): Unit
    def convertLastBehaviorToNTimes(n: Int): Unit
    def verifyAllExpectationsAreConsumed: Unit
}

class Records(context: MockContext) extends Expectations {
    val defaults = new ExpectationsImpl("defaults", context.logging)
    val expectations = new ExpectationsImpl("expectations", context.logging)

    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A] = {
        context.state.current match {
            case DefaultRecording => defaults.add(expectation, behavior)
            case _ =>
                defaults.findAndRemove(expectation.mock, expectation.method)
                expectations.add(expectation, behavior)
        }
    }

    def lastOption(): Option[RecordedBehavior] = context.state.current match {
        case DefaultRecording => defaults.lastOption
        case _                => expectations.lastOption
    }

    def convertLastBehaviorToAnyTimes() {
        context.state.current match {
            case DefaultRecording => defaults.convertLastBehaviorToAnyTimes
            case _                => expectations.convertLastBehaviorToAnyTimes
        }
    }

    def convertLastBehaviorToNTimes(n: Int) {
        context.state.current match {
            case DefaultRecording => defaults.convertLastBehaviorToNTimes(n)
            case _                => expectations.convertLastBehaviorToNTimes(n)
        }
    }

    def find(mock: Mock, method: Method, actualArgs: List[AnyRef]): Option[RecordedBehavior] =
        expectations.findAndRemove(mock, method, actualArgs).orElse(defaults.find(mock, method, actualArgs))

    def verifyAllExpectationsAreConsumed {
        expectations.verifyAllExpectationsAreConsumed
    }

    override def toString = defaults + "\n" + expectations
}

class ExpectationsImpl(name: String, logging: Logging) extends Expectations {
    val expectations = new ListBuffer[RecordedBehavior]

    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A] = {
        expectation.returnsUnit match {
            case false =>
                logging.printVerbose("Recording " + "[" + name + "] :\n  " + expectation + "\n   " + behavior)
                expectations.append(RecordedBehavior(expectation, behavior))
            case true => behavior match {
                case throws @ Throws(_) =>
                    logging.printVerbose("   " + throws)
                    expectations.remove(expectations.size - 1)
                    expectations.append(RecordedBehavior(expectation, throws))
                case nop @ Nop() =>
                    logging.printVerbose("Recording " + "[" + name + "] :\n  " + expectation)
                    expectations.append(RecordedBehavior(expectation, nop))
                case _ => throw new PebkacError("returns behavior is not allowed on mocked functions that return Unit.")
            }
        }

        expectation
    }

    def lastOption = expectations.lastOption

    def convertLastBehaviorToAnyTimes() = {
        expectations.lastOption match {
            case Some(last) =>
                expectations.remove(expectations.size - 1)
                val newBehavior = last.copy(behavior = AnyTimes(last.behavior))
                expectations.append(newBehavior)
                logging.printVerbose(newBehavior.toString)
            case None => throw new PebkacError("FOREVER WHAT!?")
        }
    }

    def convertLastBehaviorToNTimes(n: Int) = {
        expectations.lastOption match {
            case Some(last) =>
                expectations.remove(expectations.size - 1)
                val newBehavior = last.copy(behavior = Times(n, last.behavior))
                expectations.append(newBehavior)
                logging.printVerbose(newBehavior.toString)
            case None => throw new PebkacError(n + " TIMES WHAT!?")
        }
    }

    def find(mock: Mock, method: Method, actualArgs: List[AnyRef]): Option[RecordedBehavior] = {
        findTheBestOne(mock, method, actualArgs)(expectations.toList).flatMap(recordedBehavior => {
            logging.printVerbose("Replaying default " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
            Some(recordedBehavior)
        })
        //        expectations.find(expectation => (mock eq expectation.expectation.mock) && method == expectation.expectation.method && actualArgs == expectation.expectation.args).flatMap(recordedBehavior => {
        //            logging.printVerbose("Replaying default " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
        //            Some(recordedBehavior)
        //        })
    }

    def findAndRemove(mock: Mock, method: Method): Option[RecordedBehavior] = {
        expectations.zipWithIndex
            .find { case (recordedBehavior, index) => (mock eq recordedBehavior.expectation.mock) && method == recordedBehavior.expectation.method }
            .map {
                case (recordedBehavior, index) => recordedBehavior.behavior match {
                    case AnyTimes(behavior) =>
                        logging.printVerbose("Replaying anytimes " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    case Times(n, behavior) =>
                        val recordedBehavior = expectations.remove(index)
                        logging.printVerbose("Replaying times " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        val times = recordedBehavior.behavior.asInstanceOf[Times]
                        if (n > 1) expectations.append(recordedBehavior.copy(behavior = times.copy(n = times.n - 1)))
                        recordedBehavior
                    case _ => {
                        val recordedBehavior = expectations.remove(index)
                        logging.printVerbose("Replaying " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    }
                }
            }
    }
    def findAndRemove(mock: Mock, method: Method, actualArgs: List[AnyRef]): Option[RecordedBehavior] = {
        findTheBestOneIndexed(mock, method, actualArgs)(expectations.toList)
//        expectations.zipWithIndex
//            .find { case (recordedBehavior, index) => (mock eq recordedBehavior.expectation.mock) && method == recordedBehavior.expectation.method && actualArgs == recordedBehavior.expectation.args }
            .map {
                case (recordedBehavior, index) => recordedBehavior.behavior match {
                    case AnyTimes(behavior) =>
                        logging.printVerbose("Replaying anytimes " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    case Times(n, behavior) =>
                        val recordedBehavior = expectations.remove(index)
                        logging.printVerbose("Replaying times " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        val times = recordedBehavior.behavior.asInstanceOf[Times]
                        if (n > 1) expectations.append(recordedBehavior.copy(behavior = times.copy(n = times.n - 1)))
                        recordedBehavior
                    case _ => {
                        val recordedBehavior = expectations.remove(index)
                        logging.printVerbose("Replaying " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    }
                }
            }
    }

    private def findTheBestOne(mock: Mock, method: Method, actualArgs: List[AnyRef]) = { (recordedBehaviors: List[RecordedBehavior]) =>
        val matchingMethods = recordedBehaviors.filter(recordedBehavior => (mock eq recordedBehavior.expectation.mock) && method == recordedBehavior.expectation.method)
        val perfect = matchingMethods.find(_.expectation.args == actualArgs)
        val possibles = matchingMethods.filterNot { r =>
            val argsMatch = r.expectation.matchers.zip(actualArgs).map {
                case (matcher, actual) => matcher.matches(actual)
            }
            argsMatch.contains(false)
        }
        perfect.orElse(possibles.headOption).orElse(matchingMethods.headOption)
    }

    private def findTheBestOneIndexed(mock: Mock, method: Method, actualArgs: List[AnyRef]) = { (recordedBehaviors: List[RecordedBehavior]) =>
        val matchingMethods = recordedBehaviors.zipWithIndex.filter { case (recordedBehavior, index) => (mock eq recordedBehavior.expectation.mock) && method == recordedBehavior.expectation.method }
        val perfect = matchingMethods.find { case (recordedBehavior, index) => recordedBehavior.expectation.args == actualArgs }
        val possibles = matchingMethods.filterNot { case (r, index) =>
            val argsMatch = r.expectation.matchers.zip(actualArgs).map {
                case (matcher, actual) => matcher.matches(actual)
            }
            argsMatch.contains(false)
        }
        perfect.orElse(possibles.headOption).orElse(matchingMethods.headOption)
    }

    def verifyAllExpectationsAreConsumed {
        val regularExpections = expectations.toList.filterNot(_.behavior.isInstanceOf[AnyTimes])

        logging.printSystemError("LEFT OVER expectations=" + expectations)

        if (!regularExpections.isEmpty)
            throw new VerificationError("Expected methods were not invoked by the unit!\n" + regularExpections.map(_.methodName).mkString(" - ", "\n - ", ""))
    }

    override def toString = name + ":\n" + expectations
}

class LastExpectation(expectations: Records, logging: Logging) {
    private var lastExpectation: Option[Expectation[_]] = None

    def store[A](expectation: Expectation[A]) = expectation.returnsUnit match {
        case false => lastExpectation = Some(expectation)
        case true =>
            lastExpectation = Some(expectation)
            expectations.add(expectation, Nop())
    }
    def last[A]() = lastExpectation.asInstanceOf[Option[Expectation[A]]]
    def clear() = lastExpectation = None
    def isComplete(): Option[Expectation[_]] = {

        logging.printSystemError(expectations.lastOption + " AND " + lastExpectation)
        logging.printSystemError(expectations.toString)

        (expectations.lastOption, lastExpectation) match {
            case (Some(recordedBehavior), Some(expectation)) if (expectation eq recordedBehavior.expectation) => lastExpectation = None
            case _ =>
        }
        lastExpectation
    }
}

trait Logging {
    val printSystemErrors = false
    val printVerbose = false

    def printVerbose(s: String) {
        if (printVerbose)
            println(s)
    }

    def printSystemError(s: String) {
        if (printSystemErrors)
            println(s)
    }
}

trait LogSystemErrors { self: Logging =>
    override val printSystemErrors = true
}

trait LogVerbose { self: Logging =>
    override val printVerbose = true
}

trait State
case object Recording extends State
case object DefaultRecording extends State
case object Replaying extends State
case object Done extends State

class CurrentState {
    private var _state: State = Recording

    def current = _state

    def replay = current match {
        case Recording        => _state = Replaying
        case DefaultRecording => throw new StateError("Cannot call test in a defaults block")
        case Replaying        => // This is ok
        case Done             => throw new StateError("Cannot call anything after test is done :(")
    }

    def default = current match {
        case Recording        => _state = DefaultRecording
        case DefaultRecording => throw new StateError("Cannot have a defaults block inside a defaults block")
        case Replaying        => throw new StateError("Cannot have a defaults block in or after test")
        case Done             => throw new StateError("Cannot have a defaults block in or after test")
    }

    def exitDefault = current match {
        case Recording        => throw new WtfError("Recording -> exitDefault ??")
        case DefaultRecording => _state = Recording
        case Replaying        => throw new WtfError("Replaying -> exitDefault ??")
        case Done             => throw new WtfError("Done -> exitDefault ??")
    }

    def done = current match {
        case Recording        => throw new WtfError("Recording -> Done ??")
        case DefaultRecording => throw new WtfError("Default -> Done ??")
        case Replaying        => _state = Done
        case Done             => throw new WtfError("Done -> Done ??")
    }
}

class MockContext(val logging: Logging) extends Matchers {
    val state = new CurrentState
    val expectations = new Records(this)
    val currentMatchers = new RecordedMatchers(this)
    val lastExpectation = new LastExpectation(expectations, logging)
}